#include <cstring>
#include <fstream>
#include <iostream>
#include <sstream>
#include <string>
#include "divsufsort.h"

using namespace std;

string T;     // text
long long N;  // size
int *SA;      // suffix array
int *CPSS;    // array of size 2*n - contains LPnF and prevOcc entries

//// CODE BY DOMINIK KEMPA --> ////
void kkp3(unsigned char *X, int n) {
    if (n == 0) return;
    CPSS = new int[2 * n + 10];

    // Compute PSV_text and NSV_text for SA.
    int top = 0;
    for (int i = 1; i <= n + 1; ++i) {
        while (SA[top] > SA[i]) {
            int addr = (SA[top] << 1);
            CPSS[addr] = SA[top - 1];
            CPSS[addr + 1] = SA[i];
            --top;
        }
        SA[++top] = SA[i];
    }

    int plpf = 0, nlpf = 0;
    CPSS[0] = 0;
    CPSS[1] = -1;
    for (int i = 1; i < n; ++i) {
        plpf = std::max(plpf - 1, 0);
        nlpf = std::max(nlpf - 1, 0);
        int psv = CPSS[2 * i];
        int nsv = CPSS[2 * i + 1];
        if (psv != -1)
            while (i + plpf < n && X[psv + plpf] == X[i + plpf]) ++plpf;
        if (nsv != -1)
            while (i + nlpf < n && X[nsv + nlpf] == X[i + nlpf]) ++nlpf;
        if (plpf >= nlpf) {
            CPSS[2 * i] = plpf;     // len
            CPSS[2 * i + 1] = psv;  // pos
        } else {
            CPSS[2 * i] = nlpf;
            CPSS[2 * i + 1] = nsv;
        }
    }
}
//// <-- CODE BY DOMINIK KEMPA ////

void compute_lmOcc() {  // algorithm 2 stated in the paper
    for (int i = 0; i < N - 1; i++) {
        if (CPSS[2 * i] > 0) {                    // SA
            int j = CPSS[2 * i + 1];              // prevOcc
            while (CPSS[2 * j] >= CPSS[2 * i]) {  // SA
                j = CPSS[2 * j + 1];              // prevOcc
            }
            CPSS[2 * i + 1] = j;  // prevOcc
        }
    }
}

void computeLPnF_modified() {  // modified algorithm 1 stated in the paper
    for (int i = N - 1; i >= 0; i--) {
        if (CPSS[2 * i] > 0) {          // LPnF
            int j = CPSS[2 * i + 1];    // prevOcc
            if (j + CPSS[2 * i] > i) {  // LPnF
                int l = i - j;
                if (CPSS[2 * j] > l) {                  // LPnF
                    CPSS[2 * i + 1] = CPSS[2 * j + 1];  // prevOcc
                    CPSS[2 * i] = CPSS[2 * j];          // LPnF
                } else {
                    CPSS[2 * i + 1] = j;  // prevOcc
                    CPSS[2 * i] = l;      // LPnF
                }
            }
        }
    }
}

void computeLPnF() {  // algorithm 1 stated in the paper
    for (int i = N - 1; i >= 0; i--) {
        if (CPSS[2 * i] > 0) {          // LPnF
            int j = CPSS[2 * i + 1];    // prevOcc
            if (j + CPSS[2 * i] > i) {  // LPnF
                int l = i - j;
                while (CPSS[2 * j] > l) {               // LPnF
                    l = min(CPSS[2 * i], CPSS[2 * j]);  // LPnF
                    j = CPSS[2 * j + 1];                // prevOcc
                    if (j + l <= i) {
                        break;
                    } else {
                        l = i - j;
                    }
                }
                CPSS[2 * i + 1] = j;  // prevOcc
                CPSS[2 * i] = l;      // LPnF
            }
        }
    }
}

void loadFile(char *fname) {
    ifstream ifs(fname, ifstream::in);
    if (ifs.fail()) {
        cerr << "Could not open file " << fname << endl;
        exit(EXIT_FAILURE);
    }
    stringstream ss;
    ss << ifs.rdbuf();
    T = ss.str();
    ifs.close();

    N = T.length();
    if (N <= 0) exit(0);
}

int main(int argc, char *argv[]) {
    if (3 <= argc && argc <= 5) {
        loadFile(argv[1]);

        unsigned char *s = (unsigned char *)&T[0];
        SA = new int[N + 2];
        divsufsort(s, SA + 1, N);
        SA[0] = SA[N + 1] = -1;

        kkp3(s, N);

        if (strcmp(argv[2], "1") == 0) {
            compute_lmOcc();         // calculate leftmost prevOcc
            computeLPnF_modified();  // calculate LPnF values
        } else if (strcmp(argv[2], "0") == 0) {
            computeLPnF();  // calculate LPnF values from a non leftmost
                            // prevOcc-array
        } else {
            cerr << "usage: " << argv[0] << " <filepath> <with-alg2=0/1> "
                 << "[<LPnF/f-factor_path>] [<prev-occ_path>]" << endl;
            return EXIT_FAILURE;
        }

        // f-factor output
        if (4 <= argc && argc <= 5) {
            ofstream LPnF_out(argv[3], ifstream::out);
            for (int i = 0; i < N; i++) {
                LPnF_out << CPSS[2 * i] << " ";
#ifdef f_factor
                if (CPSS[2 * i] > 1) {
                    i += CPSS[2 * i] - 1;
                }
#endif
            }
            LPnF_out << endl;
            LPnF_out.close();
        }

        // prevOcc output
        if (argc == 5) {
            ofstream prevOcc_out(argv[4], ifstream::out);
            for (int i = 0; i < N; i++) {
                if (CPSS[2 * i] == 0) {
                    prevOcc_out << T[i] << " ";
                } else {
                    prevOcc_out << CPSS[2 * i + 1] << " ";
                }
#ifdef f_factor
                if (CPSS[2 * i] > 1) {
                    i += SA[i] - 1;
                }
#endif
            }
            prevOcc_out << endl;
            prevOcc_out.close();
        }

        delete[] SA;
        delete[] CPSS;
    } else {
        cerr << "usage: " << argv[0] << " <filepath> <with-alg2=0/1> "
             << "[<LPnF/f-factor_path>] [<prev-occ_path>]" << endl;
        return EXIT_FAILURE;
    }
}
