#include <algorithm>
#include <fstream>
#include <iostream>
#include <sstream>
#include <string>
#include "sdsl/rmq_support.hpp"
#include "sdsl/suffix_arrays.hpp"
#include "sdsl/suffix_array_algorithm.hpp"

using namespace std;

#ifndef SAMPLE
#define SAMPLE 1
#endif

sdsl::int_vector<8> T;                                       // reversed text
size_t N;                                                    // size
sdsl::csa_wt<sdsl::wt_huff<>, SAMPLE, 10 * 1024 * 1024> SA;  // suffix-array
sdsl::rmq_succinct_sct<false> RMQ;  // range maximum queries
ofstream LPnF_out;
ofstream prevOcc_out;

void f_factorization() {  // algortihm 3 stated in the paper
    int i = N - 1, pos, m, max;
    size_t sp, ep;
    while (i >= 0) {
        sp = 0, ep = N;
        pos = i, m = 0;
        do {
            sdsl::backward_search(SA, sp, ep, T[i], sp, ep);
            max = SA[RMQ(sp, ep)];
            if (max <= pos) break;
            m = max;
            i--;
        } while (i >= 0);
        int LPnF = pos - i;
        LPnF_out << LPnF << " ";
        if (LPnF == 0) {
            prevOcc_out << T[i] << " ";
            i--;
        } else {
            prevOcc_out << N - (m + LPnF) << " ";
        }
    }
    LPnF_out << endl;
    prevOcc_out << endl;
}

string loadFile(char* fname) {
    ifstream ifs(fname, ifstream::in);
    if (ifs.fail()) {
        cerr << "Could not open file " << fname << endl;
        exit(EXIT_FAILURE);
    }
    stringstream ss;
    ss << ifs.rdbuf();
    ifs.close();
    return ss.str();
}

int main(int argc, char** argv) {
    if (2 <= argc && argc <= 4) {
        if (3 <= argc && argc <= 4) {
            LPnF_out = ofstream(argv[2]);
        }
        if (argc == 4) {
            prevOcc_out = ofstream(argv[3]);
        }

        // reverse string
        sdsl::cache_config config(false);
        string text = loadFile(argv[1]);
        string tmp_key = sdsl::util::to_string(sdsl::util::pid()) + "_" +
                         sdsl::util::to_string(sdsl::util::id());
        string tmp_file_name = cache_file_name(tmp_key, config);
        reverse(text.begin(), text.end());
        sdsl::store_to_file(text, tmp_file_name);
        sdsl::util::clear(text);

        // construct WT
        sdsl::construct(SA, tmp_file_name, config, 1);
        sdsl::remove(tmp_file_name);
        sdsl::load_from_cache(T, sdsl::key_text_trait<8>::KEY_TEXT, config);
        sdsl::util::delete_all_files(config.file_map);
        N = T.size() - 1;

        // construct RMQ
        RMQ = sdsl::rmq_succinct_sct<false>(&SA);

        // f-factorizazion
        f_factorization();

        if (3 <= argc && argc <= 4) {
            LPnF_out.close();
        }
        if (argc == 4) {
            prevOcc_out.close();
        }
    } else {
        cerr << "usage: " << argv[0]
             << " <input-filepath> [<f-factor_path>] [<prev-occ_path>]" << endl;
        return EXIT_FAILURE;
    }
}
